import os
import mock
import google.auth.credentials
from google.cloud import firestore

from etl.execution.process import Process


if os.getenv('GAE_ENV', '').startswith('standard'):
    # production
    db = firestore.Client()
else:
    # localhost
    os.environ["FIRESTORE_DATASET"] = "test"
    os.environ["FIRESTORE_EMULATOR_HOST"] = "localhost:8888"
    os.environ["FIRESTORE_EMULATOR_HOST_PATH"] = "localhost:8888/firestore"
    os.environ["FIRESTORE_HOST"] = "http://localhost:8888"
    os.environ["FIRESTORE_PROJECT_ID"] = "boast-firebase"

    credentials = mock.Mock(spec=google.auth.credentials.Credentials)
    db = firestore.Client(project="boast-firebase", credentials=credentials)


def setup(tenant, user, fiscal_year):
    db.document(f'tenants/{tenant}').set({'debug': True})
    db.document(f'tenants/{tenant}/integrations/Github').set({'debug': True})

    db.document(f'tenants/{tenant}/integrations/Github/users/{user["name"]}').set(user)

    for config in user['configs']:
        db.document(f'tenants/{tenant}/integrations/Github/users/{user["name"]}/configs/{config["id"]}').set(config)

    db.document(f'tenants/{tenant}/integrations/Github/fiscal_years/{fiscal_year["name"]}').set(fiscal_year)
    db.document(f'tenants/{tenant}/integrations/Github/fiscal_years/{fiscal_year["name"]}/etl_jobs/1').set({
        'debug': True
    })

    ingest_path = f'tenants/{tenant}/integrations/Github/fiscal_years/{fiscal_year["name"]}/etl_jobs/1/jobs/ingest'
    db.document(ingest_path).set({
        'active': True,
        'status': 'new'
    })
    return ingest_path


if __name__ == '__main__':

    tenant = 'test'
    user = {
        'name': 'debug',
        'access_token': '14d4640066826957891dd3ce521ea3699742c4fd',
        'configs': [{
            'id': 1,
            'default_branch': 'master',
            'full_name': 'BoastCapitalLP/abi',
            'read_from': True
        }]
    }
    fiscal_year = {
        'name': 'debug FYE',
        'start_date': '2021-01-01',
        'end_date': '2021-12-31'
    }

    path = setup(tenant=tenant, user=user, fiscal_year=fiscal_year)

    Process(db=db, doc_path=path).run()



import random
import time

from etl.execution.exceptions import RetryException, SleepException
from etl.execution.threading import TimedThread
from etl.execution.state import State
from etl.execution.time import Timer
from etl.execution.config import Status


class Worker(TimedThread):

    def __init__(self, timer: Timer, event_id: str, state: State):
        TimedThread.__init__(self, timer=timer, event_id=event_id, name=self.__class__.__name__)

        self.state = state

    def get_sleep_duration(self):
        return self.state.get_sleep_duration()

    def run(self):
        self.state.set_created_at()
        self.state.set_modified_at()

        if self.state.status == Status.SLEEP:
            self.state.set_status(Status.SLEEP)

            duration = self.get_sleep_duration()

            # Wait duration / can be interrupted by timeout

            self.logger.debug(f"# SLEEP START {duration}")
            self.timer.timeout_event.wait(duration)
            self.logger.debug(f"# SLEEP END")

            # Move to WAITING_RETRY state and resume execution on next call

            self.state.set_status(Status.WAITING_RETRY)

        if self.state.status == Status.RUNNING:
            self.logger.debug(f"# RUNNING")
            try:
                self.state.set_status(Status.RUNNING)
                self.work()
                self.state.set_status(Status.DONE)
            except RetryException:
                self.state.set_status(Status.WAITING_RETRY)
            except SleepException:
                self.state.set_status(Status.WAITING_SLEEP)
            except Exception as e:
                self.state.set_status(Status.FAILED, error=e)
                raise e

    def work(self):
        raise NotImplementedError


class NoopWorker(Worker):

    def work(self):
        sleep = random.randint(0, 15)
        self.logger.debug(f"Sleep {sleep}")
        time.sleep(sleep)

        # if (random.randint(0, 9) == 1):
        #     raise Exception('Ingester Kind #1 exception')
        # if (random.randint(0, 9) == 2):
        #     raise Exception('Ingester Kind #2 exception')



API_VERSION = "1.0.0"

TIMEOUT_SECONDS = 360


class Status:
    NEW = 'new'
    RUNNING = 'running'
    FAILED = 'failed'
    DONE = 'done'
    SLEEP = 'sleep'

    WAITING_RETRY = 'waiting_retry'
    WAITING_SLEEP = 'waiting_sleep'



class ExecutionNotAllowed(Exception):

    def __init__(self):
        super().__init__('Execution Not Allowed!')


class SleepException(Exception):

    def __init__(self):
        super().__init__('Awaiting sleep!')


class RetryException(Exception):

    def __init__(self):
        super().__init__('Awaiting retry!')


class TimeoutException(RetryException):
    pass



import logging
import semantic_version

from google.cloud import firestore_v1

from etl.execution.config import API_VERSION, Status
from etl.execution.exceptions import ExecutionNotAllowed


class Event(object):
    # state: State

    def __init__(self, db, path: str):
        self.db = db
        self.event_ref = db.document(path)
        self.job_ref = self.event_ref.parent.parent

        self.ID = self.event_ref.id

        self.logger = self._init_logger()

    def _init_logger(self):
        logger = logging.getLogger(f"[{self.ID}] [{self.__class__.__name__}]")
        logger.setLevel(logging.DEBUG)

        stdout = logging.StreamHandler()
        logger.addHandler(stdout)

        stdout.setFormatter(logging.Formatter('%(asctime)s %(name)s %(message)s'))
        return logger

    def read_event_doc(self, transaction):
        return self.event_ref.get(transaction=transaction).to_dict()

    def read_state_doc(self, transaction):
        return self.job_ref.get(transaction=transaction).to_dict()

    # ------
    # Version
    # ------

    def set_api_version_error(self, transaction):
        transaction.update(self.event_ref, {
            'error': f'Incompatible api version found {self.api_version}, required {API_VERSION}'
        })

    def set_api_version(self, transaction):
        transaction.update(self.event_ref, {'api_version': API_VERSION})

    # ------
    # Handle
    # ------

    def set_handled(self, transaction, handled):
        transaction.update(self.event_ref, {'handled': handled})

    def handle(self):
        raise NotImplementedError

    # ------
    # Debug
    # ------

    def increase_invocation_count(self, transaction):
        transaction.update(self.event_ref, {'invocation_count': firestore_v1.transforms.Increment(1)})

    def increase_execution_count(self, transaction):
        transaction.update(self.event_ref, {'execution_count': firestore_v1.transforms.Increment(1)})


class ExecutionEvent(Event):

    def handle(self):

        @firestore_v1.transactional
        def handle_in_transaction(transaction: firestore_v1.Transaction):
            # IMPORTANT
            # Reads first

            event_dict = self.read_event_doc(transaction)
            state_dict = self.read_state_doc(transaction)

            api_version = event_dict.get('api_version', API_VERSION)
            event_handled = event_dict.get('handled', False)

            # IMPORTANT
            # Writes after

            # Check api version compatibility
            if semantic_version.Version(api_version).major != semantic_version.Version(API_VERSION).major:
                self.logger.debug('Incompatible api version')
                self.set_api_version_error(transaction)
                return

            if not event_handled:
                self.set_api_version(transaction)
                self.set_handled(transaction, True)

            old_status = state_dict.get('status', Status.NEW)
            new_status = old_status

            allow_execution = False

            if old_status in [Status.NEW, Status.WAITING_RETRY]:
                new_status = Status.RUNNING
                transaction.update(self.job_ref, {'status': new_status})
                allow_execution = True

            if old_status == Status.WAITING_SLEEP:
                new_status = Status.SLEEP
                transaction.update(self.job_ref, {'status': new_status})
                allow_execution = True

            invocation = {
                'old_status': old_status,
                'new_status': new_status,
                'allow_execution': allow_execution,
            }
            transaction.update(self.job_ref, {'invocations': firestore_v1.transforms.ArrayUnion([invocation])})

            if allow_execution:
                self.increase_execution_count(transaction)

            self.increase_invocation_count(transaction)

            self.logger.debug(invocation)

            return allow_execution

        allow_execution = handle_in_transaction(transaction=firestore_v1.Transaction(client=self.db))

        if not allow_execution:
            raise ExecutionNotAllowed


class NotifyEvent(Event):
    pass

